# Generated by Django 2.2.6 on 2019-10-28 18:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_repository_pulp_type'),
        ('file', '0004_filefilesystemexporter'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileRepository',
            fields=[
                ('repository_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='file_filerepository', serialize=False, to='core.Repository')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.repository',),
        ),
    ]



import logging
import requests
import tempfile
import time

from pulp_smash import config

from .utils import measureit, urljoin

cfg = config.get_config()
BASE_ADDR = cfg.get_base_url()


def get(url, params={}):
    """Wrapper around requests.get with some simplification in our case."""
    url = BASE_ADDR + url

    r = requests.get(url=url, params=params)
    r.raise_for_status()
    data = r.json()
    return data


def get_results(url, params={}):
    """Wrapper around requests.get with some simplification in our case."""
    out = []
    params["limit"] = 100
    params["offset"] = 0
    while True:
        data = get(url, params)
        out += data["results"]
        params["offset"] += 100
        if data["next"] is None:
            break
    return out


def post(url, data):
    """Wrapper around requests.post with some simplification in our case."""
    url = BASE_ADDR + url

    r = requests.post(url=url, json=data)
    r.raise_for_status()
    return r.json()


def download(base_url, file_name, file_size):
    """Downlad file with expected size and drop it."""
    with tempfile.TemporaryFile() as downloaded_file:
        full_url = urljoin(base_url, file_name)
        duration, response = measureit(requests.get, full_url)
        response.raise_for_status()
        downloaded_file.write(response.content)
        assert downloaded_file.tell() == file_size
        return duration


def wait_for_tasks(tasks, timeout=None):
    """
    Wait for tasks to finish.

    Returns task info. If we time out, list of None is returned.
    """
    start = time.time()
    out = []
    step = 3
    for t in tasks:
        while True:
            if timeout is not None:
                now = time.time()
                if now >= start + timeout:
                    raise Exception("Task %s timeouted" % t)
            response = get(t)
            logging.debug("Task status is '%s', full response %s" % (response["state"], response))
            if response["state"] in ("failed", "cancelled", "completed"):
                out.append(response)
                break
            else:
                time.sleep(step)
    return out



# Generated by Django 2.2.6 on 2019-10-14 17:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('file', '0002_file_related_names'),
    ]

    operations = [
        migrations.AlterField(
            model_name='filecontent',
            name='relative_path',
            field=models.TextField(),
        ),
    ]



# Generated by Django 2.2.17 on 2020-12-05 18:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('file', '0007_filefilesystemexporter'),
    ]

    operations = [
        migrations.AddField(
            model_name='filepublication',
            name='manifest',
            field=models.TextField(default='PULP_MANIFEST'),
            preserve_default=False,
        ),
    ]



from django.test import TestCase


class TestNothing(TestCase):
    """Test Nothing (placeholder)."""

    def test_nothing_at_all(self):
        """Test that the tests are running and that's it."""
        self.assertTrue(True)



# coding=utf-8
"""Tests that sync file plugin repositories."""
import unittest

from pulp_smash import config
from pulp_smash.pulp3.bindings import monitor_task, PulpTaskError
from pulp_smash.pulp3.utils import gen_repo, get_added_content_summary, get_content_summary

from pulp_file.tests.functional.constants import (
    FILE2_FIXTURE_MANIFEST_URL,
    FILE_FIXTURE_SUMMARY,
    FILE_INVALID_MANIFEST_URL,
)
from pulp_file.tests.functional.utils import gen_file_client, gen_file_remote
from pulp_file.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_file import (
    RepositoriesFileApi,
    RepositorySyncURL,
    RemotesFileApi,
)


class BasicSyncTestCase(unittest.TestCase):
    """Sync a repository with the file plugin."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = gen_file_client()

    def test_sync(self):
        """Sync repositories with the file plugin.

        In order to sync a repository a remote has to be associated within
        this repository. When a repository is created this version field is set
        as None. After a sync the repository version is updated.

        Do the following:

        1. Create a repository, and a remote.
        2. Assert that repository version is None.
        3. Sync the remote.
        4. Assert that repository version is not None.
        5. Assert that the correct number of units were added and are present
           in the repo.
        6. Sync the remote one more time.
        7. Assert that repository version is different from the previous one.
        8. Assert that the same number of are present and that no units were
           added.
        """
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote()
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        self.assertIsNotNone(repo.latest_version_href)
        self.assertDictEqual(get_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)
        self.assertDictEqual(get_added_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)

        # Sync the repository again.
        latest_version_href = repo.latest_version_href
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        self.assertEqual(latest_version_href, repo.latest_version_href)
        self.assertDictEqual(get_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)


class SyncInvalidTestCase(unittest.TestCase):
    """Sync a repository with a given url on the remote."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_file_client()

    def test_invalid_url(self):
        """Sync a repository using a remote url that does not exist.

        Test that we get a task failure. See :meth:`do_test`.
        """
        with self.assertRaises(PulpTaskError) as cm:
            task = self.do_test("http://i-am-an-invalid-url.com/invalid/")
        task = cm.exception.task.to_dict()
        self.assertIsNotNone(task["error"]["description"])

    def test_invalid_file(self):
        """Sync a repository using an invalid file repository.

        Assert that an exception is raised, and that error message has
        keywords related to the reason of the failure. See :meth:`do_test`.
        """
        with self.assertRaises(PulpTaskError) as cm:
            task = self.do_test(FILE_INVALID_MANIFEST_URL)
        task = cm.exception.task.to_dict()
        for key in ("checksum", "failed"):
            self.assertIn(key, task["error"]["description"])

    def do_test(self, url):
        """Sync a repository given ``url`` on the remote."""
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote(url=url)
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        return monitor_task(sync_response.task)


class SyncDuplicateFileRepoTestCase(unittest.TestCase):
    """Sync multiple remotes containing duplicate files."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_file_client()

    def test_duplicate_file_sync(self):
        """Sync a repository with remotes containing same file names.

        This test does the following.

        1. Create a repository in pulp.
        2. Create two remotes containing the same file.
        3. Check whether the created repo has only one copy of the file.

        This test targets the following issue:

        `Pulp #4738 <https://pulp.plan.io/issues/4738>`_
        """
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)

        # Step 1
        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        # Step 2
        remote = remote_api.create(gen_file_remote())
        self.addCleanup(remote_api.delete, remote.pulp_href)
        remote2 = remote_api.create(gen_file_remote(url=FILE2_FIXTURE_MANIFEST_URL))

        self.addCleanup(remote_api.delete, remote2.pulp_href)
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)
        self.assertDictEqual(get_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)
        self.assertDictEqual(get_added_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)

        repository_sync_data = RepositorySyncURL(remote=remote2.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)
        self.assertDictEqual(get_added_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)



# coding=utf-8
"""Tests whether Pulp handles PULP_MANIFEST information."""
import csv
import requests
import unittest
from urllib.parse import urljoin

from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import gen_distribution, gen_repo

from pulp_file.tests.functional.constants import FILE_FIXTURE_COUNT
from pulp_file.tests.functional.utils import gen_file_client, gen_file_remote
from pulp_file.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_file import (
    DistributionsFileApi,
    PublicationsFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
    RemotesFileApi,
    FileFilePublication,
)


class AccessingPublishedDataTestCase(unittest.TestCase):
    """Assert that an HTTP error is not raised when accessing published data.

    This test targets the following issue:

    * `Pulp #4519 https://pulp.plan.io/issues/4519`_
    """

    @classmethod
    def setUpClass(cls):
        """Define class-wide variable."""
        cls.client = gen_file_client()

    def test_access_error(self):
        """HTTP error is not raised when accessing published data."""
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)
        publications = PublicationsFileApi(self.client)
        distributions = DistributionsFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        remote = remote_api.create(gen_file_remote())
        self.addCleanup(remote_api.delete, remote.pulp_href)

        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        publish_data = FileFilePublication(repository=repo.pulp_href)
        publish_response = publications.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        self.addCleanup(publications.delete, publication_href)

        body = gen_distribution()
        body["publication"] = publication_href

        distribution_response = distributions.create(body)
        created_resources = monitor_task(distribution_response.task).created_resources
        distribution = distributions.read(created_resources[0])
        self.addCleanup(distributions.delete, distribution.pulp_href)

        pulp_manifest = parse_pulp_manifest(
            self.download_pulp_manifest(distribution.to_dict(), "PULP_MANIFEST")
        )

        self.assertEqual(len(pulp_manifest), FILE_FIXTURE_COUNT, pulp_manifest)

    def download_pulp_manifest(self, distribution, unit_path):
        """Download pulp manifest."""
        unit_url = urljoin(distribution["base_url"] + "/", unit_path)
        return requests.get(unit_url)


def parse_pulp_manifest(pulp_manifest):
    """Parse pulp manifest."""
    return list(csv.DictReader(pulp_manifest.text.splitlines(), ("name", "checksum", "size")))



import logging
import random
import time
import string
import requests


def get_random_string():
    """Return random string."""
    return "".join(random.choice(string.ascii_lowercase) for i in range(5))


def urljoin(*args):
    """This sucks, but works. Better ways welcome."""
    return "/".join([i.lstrip("/").rstrip("/") for i in args])


def measureit(func, *args, **kwargs):
    """Measure execution time of passed function."""
    logging.debug("Measuring duration of %s %s %s" % (func.__name__, args, kwargs))
    before = time.clock()
    out = func(*args, **kwargs)
    after = time.clock()
    return after - before, out


def parse_pulp_manifest(url):
    """Parse pulp manifest."""
    response = requests.get(url)
    response.text.split("\n")
    data = [i.strip().split(",") for i in response.text.split("\n")]
    return [(i[0], i[1], int(i[2])) for i in data if i != [""]]



from uuid import uuid4

from django.test import TestCase
from rest_framework.serializers import ValidationError as DRFValidationError

from pulpcore.constants import API_ROOT
from pulp_file.app import models, viewsets


class TestGetResource(TestCase):
    """
    Test NamedViewSet from core using some detail endpoints provided by file.

    Note: This is really a pulpcore test, but we can't test it from pulpcore.
    """

    def test_no_errors(self):
        """
        Tests that get_resource() properly resolves a valid URI and returns the correct resource.
        """
        repo = models.FileRepository.objects.create(name="foo")
        viewset = viewsets.FileRepositoryViewSet()
        resource = viewset.get_resource(
            "/{api_root}repositories/file/file/{pk}/".format(api_root=API_ROOT, pk=repo.pk),
            models.FileRepository,
        )
        self.assertEquals(repo, resource)

    def test_multiple_matches(self):
        """
        Tests that get_resource() raises a ValidationError if you attempt to use a list endpoint.
        """
        models.FileRepository.objects.create(name="foo")
        models.FileRepository.objects.create(name="foo2")
        viewset = viewsets.FileRepositoryViewSet()

        with self.assertRaises(DRFValidationError):
            # matches all repositories
            viewset.get_resource(
                "/{api_root}repositories/file/file/".format(api_root=API_ROOT),
                models.FileRepository,
            )

    def test_invalid_uri(self):
        """
        Tests that get_resource raises a ValidationError if you attempt to use an invalid URI.
        """
        viewset = viewsets.FileRepositoryViewSet()

        with self.assertRaises(DRFValidationError):
            viewset.get_resource("/pulp/api/v2/nonexistent/", models.FileRepository)

    def test_resource_does_not_exist(self):
        """
        Tests that get_resource() raises a ValidationError if you use a URI for a resource that
        does not exist.
        """
        pk = uuid4()
        viewset = viewsets.FileRepositoryViewSet()

        with self.assertRaises(DRFValidationError):
            viewset.get_resource(
                "/{api_root}repositories/file/file/{pk}/".format(api_root=API_ROOT, pk=pk),
                models.FileRepository,
            )

    def test_resource_with_field_error(self):
        """
        Tests that get_resource() raises a ValidationError if you use a URI that is not a valid
        model.
        """
        repo = models.FileRepository.objects.create(name="foo")
        viewset = viewsets.FileRepositoryViewSet()

        with self.assertRaises(DRFValidationError):
            # has no repo versions yet
            viewset.get_resource(
                "/{api_root}repositories/file/file/{pk}/versions/1/".format(
                    api_root=API_ROOT, pk=repo.pk
                ),
                models.FileRepository,
            )



# coding=utf-8
"""Utilities for tests for the file plugin."""
from functools import partial
import requests
from unittest import SkipTest
from tempfile import NamedTemporaryFile

from pulp_smash import api, config, selectors, utils
from pulp_smash.pulp3.utils import (
    gen_remote,
    gen_repo,
    get_content,
    require_pulp_3,
    require_pulp_plugins,
    sync,
)

from pulp_file.tests.functional.constants import (
    FILE_CONTENT_NAME,
    FILE_CONTENT_PATH,
    FILE_FIXTURE_MANIFEST_URL,
    FILE_PUBLICATION_PATH,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
    FILE_URL,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ArtifactsApi,
    TasksApi,
)
from pulpcore.client.pulp_file import ApiClient as FileApiClient


cfg = config.get_config()
configuration = cfg.get_bindings_config()


def set_up_module():
    """Skip tests Pulp 3 isn't under test or if pulp-file isn't installed."""
    require_pulp_3(SkipTest)
    require_pulp_plugins({"file"}, SkipTest)


def gen_file_client():
    """Return an OBJECT for file client."""
    return FileApiClient(configuration)


def gen_file_remote(url=FILE_FIXTURE_MANIFEST_URL, **kwargs):
    """Return a semi-random dict for use in creating a file Remote.

    :param url: The URL of an external content source.
    """
    return gen_remote(url, **kwargs)


def get_file_content_paths(repo, version_href=None):
    """Return the relative path of content units present in a file repository.

    :param repo: A dict of information about the repository.
    :param version_href: The repository version to read.
    :returns: A list with the paths of units present in a given repository.
    """
    # The "relative_path" is actually a file path and name
    return [
        content_unit["relative_path"]
        for content_unit in get_content(repo, version_href)[FILE_CONTENT_NAME]
    ]


def gen_file_content_attrs(artifact):
    """Generate a dict with content unit attributes.

    :param artifact: A dict of info about the artifact.
    :returns: A semi-random dict for use in creating a content unit.
    """
    return {"artifact": artifact["pulp_href"], "relative_path": utils.uuid4()}


def gen_file_content_upload_attrs():
    """Generate a dict with content unit attributes without artifact for upload.

    :param artifact: A dict of info about the artifact.
    :returns: A semi-random dict for use in creating a content unit.
    """
    return {"relative_path": utils.uuid4()}


def populate_pulp(cfg, url=FILE_FIXTURE_MANIFEST_URL):
    """Add file contents to Pulp.

    :param pulp_smash.config.PulpSmashConfig: Information about a Pulp application.
    :param url: The URL to a file repository's ``PULP_MANIFEST`` file. Defaults to
        :data:`pulp_smash.constants.FILE_FIXTURE_MANIFEST_URL` + ``PULP_MANIFEST``.
    :returns: A list of dicts, where each dict describes one file content in Pulp.
    """
    client = api.Client(cfg, api.json_handler)
    remote = {}
    repo = {}
    try:
        remote.update(client.post(FILE_REMOTE_PATH, gen_remote(url)))
        repo.update(client.post(FILE_REPO_PATH, gen_repo()))
        sync(cfg, remote, repo)
    finally:
        if remote:
            client.delete(remote["pulp_href"])
        if repo:
            client.delete(repo["pulp_href"])
    return client.get(FILE_CONTENT_PATH)["results"]


def create_file_publication(cfg, repo, version_href=None):
    """Create a file publication.

    :param pulp_smash.config.PulpSmashConfig cfg: Information about the Pulp
        host.
    :param repo: A dict of information about the repository.
    :param version_href: A href for the repo version to be published.
    :returns: A publication. A dict of information about the just created
        publication.
    """
    if version_href:
        body = {"repository_version": version_href}
    else:
        body = {"repository": repo["pulp_href"]}

    client = api.Client(cfg, api.json_handler)
    call_report = client.post(FILE_PUBLICATION_PATH, body)
    tasks = tuple(api.poll_spawned_tasks(cfg, call_report))
    return client.get(tasks[-1]["created_resources"][0])


skip_if = partial(selectors.skip_if, exc=SkipTest)  # pylint:disable=invalid-name
"""The ``@skip_if`` decorator, customized for unittest.

:func:`pulp_smash.selectors.skip_if` is test runner agnostic. This function is
identical, except that ``exc`` has been set to ``unittest.SkipTest``.
"""

core_client = CoreApiClient(configuration)
tasks = TasksApi(core_client)


def gen_artifact(url=FILE_URL, file=None):
    """Creates an artifact."""
    if not file:
        response = requests.get(url)
        with NamedTemporaryFile() as temp_file:
            temp_file.write(response.content)
            return ArtifactsApi(core_client).create(file=temp_file.name).to_dict()

    return ArtifactsApi(core_client).create(file=file).to_dict()



from pulpcore.plugin import PulpPluginAppConfig


class PulpFilePluginAppConfig(PulpPluginAppConfig):
    """
    Entry point for pulp_file plugin.
    """

    name = "pulp_file.app"
    label = "file"
    version = "1.6.0.post2"



import uuid
from unittest.mock import patch

from django.core.files.storage import default_storage as storage
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from pulpcore.plugin.models import Artifact, ContentArtifact, RepositoryVersion
from pulpcore.app.models.task import Task

from pulp_file.app.models import FileContent, FileRepository


class RepositoryVersionCRUDTestCase(TestCase):
    """Test RepositoryVersion CRUD and content association."""

    def setUp(self):
        """Create Artifact, Content, ContentArtifact, and Repository."""
        artifact = Artifact.objects.create(
            sha1="cf6121b0425c2f2e3a2fcfe6f402d59730eb5661",
            sha224="9a6297eb28d91fad5277c0833856031d0e940432ad807658bd2b60f4",
            sha256="c8ddb3dcf8da48278d57b0b94486832c66a8835316ccf7ca39e143cbfeb9184f",
            sha384=(
                "53a8a0cebcb7780ed7624790c9d9a4d09ba74b47270d397f5ed7bc1c46777a0f"
                "be362aaf2bbe7f0966a350a12d76e28d"
            ),
            # noqa
            sha512=(
                "a94a65f19b864d184a2a5e07fa29766f08c6d49b6f624b3dd3a36a9826"
                "7b9137d9c35040b3e105448a869c23c2aec04c9e064e3555295c1b8de6515eed4da27d"
            ),
            # noqa
            size=1024,
            file=SimpleUploadedFile("test_filename", b"test content"),
        )
        artifact.save()
        self.content = FileContent.objects.create()
        self.content.save()
        artifact_file = storage.open(artifact.file.name)
        self.content_artifact = ContentArtifact.objects.create(
            artifact=artifact, content=self.content, relative_path=artifact_file.name
        )
        self.content_artifact.save()
        self.repository = FileRepository.objects.create(name="foo")
        self.repository.save()
        self.task = Task.objects.create(
            state="Completed", name="test-task", _resource_job_id=uuid.uuid4()
        )
        self.task.save()

    @patch("pulpcore.app.models.task.get_current_job")
    def test_create_repository_version(self, mock_task):
        """Test creating a RepositoryVersion."""
        mock_task.return_value.id = self.task.pk
        with self.repository.new_version() as new_version:
            new_version.add_content(FileContent.objects.filter(pk=self.content.pk))
        self.assertTrue(RepositoryVersion.objects.filter().exists())

    def test_remove_repository_version(self):
        """Test deleting a RepositoryVersion."""
        RepositoryVersion.objects.filter().delete()
        self.assertFalse(RepositoryVersion.objects.filter().exists())



import datetime

DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"


def parse_date_from_string(s, parse_format="%Y-%m-%dT%H:%M:%S.%fZ"):
    """Parse string to datetime object.
    :param s: str like '2018-11-18T21:03:32.493697Z'
    :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
    :return: datetime.datetime
    """
    return datetime.datetime.strptime(s, parse_format)


def tasks_table(tasks, performance_task_name):
    """Return overview of tasks."""
    out = []
    for t in tasks:
        created_at = parse_date_from_string(t["pulp_created"])
        started_at = parse_date_from_string(t["started_at"])
        finished_at = parse_date_from_string(t["finished_at"])
        task_duration = finished_at - started_at
        waiting_time = started_at - created_at
        out.append(
            "\n-> {task_name} => Waiting time (s): {wait} | Service time (s): {service}".format(
                task_name=performance_task_name,
                wait=waiting_time.total_seconds(),
                service=task_duration.total_seconds(),
            )
        )
    return "\n".join(out)


def print_fmt_experiment_time(label, start, end):
    """Print formatted label and experiment time."""
    print("\n-> {} => Experiment time (s): {}".format(label, (end - start).total_seconds()))


def report_tasks_stats(performance_task_name, tasks):
    """Print out basic stats about received tasks."""
    print(tasks_table(tasks, performance_task_name))



# Generated by Django 2.2.6 on 2020-03-13 16:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('file', '0005_filerepository'),
    ]

    operations = [
        migrations.DeleteModel(
            name='FileFileSystemExporter',
        ),
    ]



# Generated by Django 2.2.6 on 2019-11-04 20:00

from django.db import migrations

# Note: This migration was throwing exceptions so all of the fields/etc were removed.
# The table gets deleted later in 0006_delete_filefilesystemexporter.py.


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0012_auto_20191104_2000'),
        ('file', '0003_auto_20191014_1721'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileFileSystemExporter',
            fields=[],
        ),
    ]



import logging

from gettext import gettext as _

from django.core.files import File

from pulpcore.plugin.models import RepositoryVersion, PublishedMetadata, RemoteArtifact
from pulpcore.plugin.tasking import WorkingDirectory

from pulp_file.app.models import FileContent, FilePublication
from pulp_file.manifest import Entry, Manifest


log = logging.getLogger(__name__)


def publish(manifest, repository_version_pk):
    """
    Create a Publication based on a RepositoryVersion.

    Args:
        manifest (str): Filename to use for manifest file.
        repository_version_pk (str): Create a publication from this repository version.

    """
    repo_version = RepositoryVersion.objects.get(pk=repository_version_pk)

    log.info(
        _("Publishing: repository={repo}, version={ver}, manifest={manifest}").format(
            repo=repo_version.repository.name, ver=repo_version.number, manifest=manifest
        )
    )

    with WorkingDirectory():
        with FilePublication.create(repo_version, pass_through=True) as publication:
            publication.manifest = manifest
            manifest = Manifest(manifest)
            manifest.write(populate(publication))
            PublishedMetadata.create_from_file(
                file=File(open(manifest.relative_path, "rb")), publication=publication
            )

    log.info(_("Publication: {publication} created").format(publication=publication.pk))


def populate(publication):
    """
    Populate a publication.

    Create published artifacts and yield a Manifest Entry for each.

    Args:
        publication (pulpcore.plugin.models.Publication): A Publication to populate.

    Yields:
        Entry: Each manifest entry.

    """

    def find_artifact():
        _artifact = content_artifact.artifact
        if not _artifact:
            _artifact = RemoteArtifact.objects.filter(content_artifact=content_artifact).first()
        return _artifact

    for content in FileContent.objects.filter(
        pk__in=publication.repository_version.content
    ).order_by("-pulp_created"):
        for content_artifact in content.contentartifact_set.all():
            artifact = find_artifact()
            entry = Entry(
                relative_path=content_artifact.relative_path,
                digest=artifact.sha256,
                size=artifact.size,
            )
            yield entry



import datetime
import multiprocessing

from collections import namedtuple
from unittest import TestCase

from pulp_file.tests.functional.constants import FILE_PERFORMANCE_FIXTURE_URL

from .pulpperf import interact
from .pulpperf import utils
from .pulpperf import reporting

Args = namedtuple("Arguments", "limit processes repositories")


class PerformanceTestCase(TestCase):
    """
    Test case for running performance tests.

    The tests are ported from https://github.com/redhat-performance/pulpperf.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        super().setUpClass()

        cls.args = Args(limit=100, processes=1, repositories=[FILE_PERFORMANCE_FIXTURE_URL])
        cls.data = []

    def test_01_sync_repository(self):
        """Measure time of synchronization."""
        for r in self.args.repositories:
            self.data.append({"remote_url": r})

        for r in self.data:
            if r["remote_url"] not in self.args.repositories:
                continue
            r["repository_name"] = utils.get_random_string()
            r["repository_href"] = create_repo(r["repository_name"])
            r["remote_name"] = utils.get_random_string()
            r["remote_href"] = create_remote(r["remote_name"], r["remote_url"])

        tasks = []
        for r in self.data:
            if r["remote_url"] not in self.args.repositories:
                continue
            task = start_sync(r["repository_href"], r["remote_href"])
            tasks.append(task)

        results = interact.wait_for_tasks(tasks)
        reporting.report_tasks_stats("Sync tasks", results)

    def test_02_resync_repository(self):
        """Measure time of resynchronization."""
        tasks = []
        for r in self.data:
            task = start_sync(r["repository_href"], r["remote_href"])
            tasks.append(task)

        results = interact.wait_for_tasks(tasks)
        reporting.report_tasks_stats("Resync tasks", results)

    def test_03_publish_repository(self):
        """Measure time of repository publishing."""
        tasks = []
        for r in self.data:
            task = create_publication(r["repository_href"])
            tasks.append(task)

        results = interact.wait_for_tasks(tasks)
        reporting.report_tasks_stats("Publication tasks", results)

        for i in range(len(results)):
            self.data[i]["publication_href"] = results[i]["created_resources"][0]
            self.data[i]["repository_version_href"] = interact.get(
                self.data[i]["publication_href"]
            )["repository_version"]

        tasks = []
        for r in self.data:
            r["distribution_name"] = utils.get_random_string()
            r["distribution_base_path"] = utils.get_random_string()
            task = create_distribution(
                r["distribution_name"], r["distribution_base_path"], r["publication_href"]
            )
            tasks.append(task)

        results = interact.wait_for_tasks(tasks)
        reporting.report_tasks_stats("Distribution tasks", results)

        for i in range(len(results)):
            self.data[i]["distribution_href"] = results[i]["created_resources"][0]
            self.data[i]["download_base_url"] = interact.get(self.data[i]["distribution_href"])[
                "base_url"
            ]

    def test_04_download_repo(self):
        """Measure time of repository downloading."""
        before = datetime.datetime.utcnow()

        for r in self.data:
            params = []
            pulp_manifest = utils.parse_pulp_manifest(r["remote_url"] + "PULP_MANIFEST")
            for f, _, s in pulp_manifest[: self.args.limit]:
                params.append((r["download_base_url"], f, s))
            with multiprocessing.Pool(processes=self.args.processes) as pool:
                pool.starmap(interact.download, params)

        after = datetime.datetime.utcnow()
        reporting.print_fmt_experiment_time("Repository download", before, after)

    def test_05_list_content(self):
        """Measure time of inspecting the repository content."""
        before = datetime.datetime.utcnow()

        durations_list = []
        for r in self.data:
            duration, content = utils.measureit(
                list_units_in_repo_ver, r["repository_version_href"]
            )
            durations_list.append(duration)

            params = []
            for c in content[: self.args.limit]:
                url = c.get("pulp_href")
                params.append((inspect_content, url))
            with multiprocessing.Pool(processes=self.args.processes) as pool:
                pool.starmap(utils.measureit, params)
        after = datetime.datetime.utcnow()
        reporting.print_fmt_experiment_time("Content inspection", before, after)

    def test_06_repo_version(self):
        """Measure time of repository cloning."""
        for r in self.data:
            r["repository_clone1_name"] = utils.get_random_string()
            r["repository_clone1_href"] = create_repo(r["repository_clone1_name"])
            r["repository_clone2_name"] = utils.get_random_string()
            r["repository_clone2_href"] = create_repo(r["repository_clone2_name"])

        tasks = []
        for r in self.data:
            task = create_repo_version_base_version(
                r["repository_clone1_href"], r["repository_version_href"]
            )
            tasks.append(task)

        results = interact.wait_for_tasks(tasks)
        reporting.report_tasks_stats("Version clone with base_version tasks", results)

        hrefs = [i["pulp_href"] for i in list_units_in_repo_ver(r["repository_version_href"])]

        tasks = []
        for r in self.data:
            task = create_repo_version_add_content_units(r["repository_clone2_href"], hrefs)
            tasks.append(task)

        results = interact.wait_for_tasks(tasks)
        reporting.report_tasks_stats("Version clone with add_content_units tasks", results)


def create_repo(name):
    """Create repository"""
    return interact.post("/pulp/api/v3/repositories/file/file/", data={"name": name})["pulp_href"]


def create_remote(name, url):
    """Create remote"""
    return interact.post(
        "/pulp/api/v3/remotes/file/file/", data={"name": name, "url": url + "PULP_MANIFEST"}
    )["pulp_href"]


def start_sync(repo, remote):
    """Start sync of the remote into the repository, return task"""
    return interact.post(repo + "sync/", data={"remote": remote, "mirror": False})["task"]


def create_publication(repo):
    """Start publication of the repository, return task"""
    return interact.post("/pulp/api/v3/publications/file/file/", data={"repository": repo})["task"]


def create_distribution(name, base_path, pub):
    """Start distribution of the repository version, return task"""
    return interact.post(
        "/pulp/api/v3/distributions/file/file/",
        data={"name": name, "base_path": base_path, "publication": pub},
    )["task"]


def list_units_in_repo_ver(repo_ver):
    """List the file content with all the fields"""
    return interact.get_results(
        "/pulp/api/v3/content/file/files/", params={"repository_version": repo_ver}
    )


def inspect_content(href):
    """Inspect a file content using href"""
    return interact.get(href)


def create_repo_version_base_version(repo, ver):
    """Create repository version based on different existing version"""
    return interact.post(repo + "modify/", data={"base_version": ver})["task"]


def create_repo_version_add_content_units(repo, hrefs):
    """Create repository version based on list of hrefs"""
    return interact.post(repo + "modify/", data={"add_content_units": hrefs})["task"]



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.plugin import models
from pulpcore.plugin.serializers import (
    ContentChecksumSerializer,
    DetailRelatedField,
    FilesystemExporterSerializer,
    PublicationDistributionSerializer,
    PublicationSerializer,
    RemoteSerializer,
    RepositorySerializer,
    SingleArtifactContentUploadSerializer,
)

from pulp_file.app.models import (
    FileContent,
    FileDistribution,
    FileFilesystemExporter,
    FileRemote,
    FileRepository,
    FilePublication,
)


class FileContentSerializer(SingleArtifactContentUploadSerializer, ContentChecksumSerializer):
    """
    Serializer for File Content.
    """

    def deferred_validate(self, data):
        """Validate the FileContent data."""
        data = super().deferred_validate(data)

        data["digest"] = data["artifact"].sha256

        content = FileContent.objects.filter(
            digest=data["digest"], relative_path=data["relative_path"]
        )

        if content.exists():
            raise serializers.ValidationError(
                _(
                    "There is already a file content with relative path '{path}' and digest "
                    "'{digest}'."
                ).format(path=data["relative_path"], digest=data["digest"])
            )

        return data

    class Meta:
        fields = (
            SingleArtifactContentUploadSerializer.Meta.fields
            + ContentChecksumSerializer.Meta.fields
        )
        model = FileContent


class FileRepositorySerializer(RepositorySerializer):
    """
    Serializer for File Repositories.
    """

    class Meta:
        fields = RepositorySerializer.Meta.fields
        model = FileRepository


class FileRemoteSerializer(RemoteSerializer):
    """
    Serializer for File Remotes.
    """

    policy = serializers.ChoiceField(
        help_text="The policy to use when downloading content. The possible values include: "
        "'immediate', 'on_demand', and 'streamed'. 'immediate' is the default.",
        choices=models.Remote.POLICY_CHOICES,
        default=models.Remote.IMMEDIATE,
    )

    class Meta:
        fields = RemoteSerializer.Meta.fields
        model = FileRemote


class FilePublicationSerializer(PublicationSerializer):
    """
    Serializer for File Publications.
    """

    distributions = DetailRelatedField(
        help_text=_("This publication is currently hosted as defined by these distributions."),
        source="file_filedistribution",
        view_name="filedistributions-detail",
        many=True,
        read_only=True,
    )
    manifest = serializers.CharField(
        help_text=_("Filename to use for manifest file containing metadata for all the files."),
        default="PULP_MANIFEST",
    )

    class Meta:
        model = FilePublication
        fields = PublicationSerializer.Meta.fields + ("distributions", "manifest")


class FileDistributionSerializer(PublicationDistributionSerializer):
    """
    Serializer for File Distributions.
    """

    class Meta:
        fields = PublicationDistributionSerializer.Meta.fields
        model = FileDistribution


class FileFilesystemExporterSerializer(FilesystemExporterSerializer):
    """
    Serializer for File file system exporters.
    """

    class Meta:
        fields = FilesystemExporterSerializer.Meta.fields
        model = FileFilesystemExporter



from logging import getLogger

from django.db import models

from pulpcore.plugin.models import (
    Content,
    FilesystemExporter,
    Publication,
    PublicationDistribution,
    Remote,
    Repository,
)
from pulpcore.plugin.publication_utils import validate_publication_paths
from pulpcore.plugin.repo_version_utils import remove_duplicates, validate_repo_version


log = getLogger(__name__)


class FileContent(Content):
    """
    The "file" content type.

    Content of this type represents a single file uniquely
    identified by path and SHA256 digest.

    Fields:
        relative_path (str): The file relative path.
        digest (str): The SHA256 HEX digest.
    """

    TYPE = "file"
    repo_key_fields = ("relative_path",)

    relative_path = models.TextField(null=False)
    digest = models.CharField(max_length=64, null=False)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("relative_path", "digest")


class FileRemote(Remote):
    """
    Remote for "file" content.
    """

    TYPE = "file"

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"


class FileRepository(Repository):
    """
    The "file" repository type.
    """

    TYPE = "file"
    CONTENT_TYPES = [FileContent]
    REMOTE_TYPES = [FileRemote]

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"

    def finalize_new_version(self, new_version):
        """
        Finalize and validate the new repository version.

        Ensure no added content contains the same `relative_path` as other content and relative
        paths don't overlap.

        Args:
            new_version (pulpcore.app.models.RepositoryVersion): The incomplete RepositoryVersion to
                finalize.

        """
        remove_duplicates(new_version)
        validate_repo_version(new_version)


class FilePublication(Publication):
    """
    Publication for 'file' content.
    """

    TYPE = "file"

    manifest = models.TextField()

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"

    def finalize_new_publication(self):
        """
        Validate that artifact paths don't overlap.
        """
        validate_publication_paths(self)


class FileDistribution(PublicationDistribution):
    """
    Distribution for 'file' content.
    """

    TYPE = "file"

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"


class FileFilesystemExporter(FilesystemExporter):
    """
    File system exporter for 'file' content.
    """

    TYPE = "filesystem"

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"



from collections import namedtuple

from gettext import gettext as _


Line = namedtuple("Line", ("number", "content"))


class Entry:
    """
    Manifest entry.

    Format: <relative_path>,<digest>,<size>.
    Lines beginning with `#` are ignored.

    Attributes:
        relative_path (str): A relative path.
        digest (str): The file sha256 hex digest.
        size (int): The file size in bytes.

    """

    def __init__(self, relative_path, size, digest):
        """
        Create a new Entry.

        Args:
            relative_path (str): A relative path.
            digest (str): The file sha256 hex digest.
            size (int): The file size in bytes.

        """
        self.relative_path = relative_path
        self.digest = digest
        self.size = size

    @staticmethod
    def parse(line):
        """
        Parse the specified line from the manifest into an Entry.

        Args:
            line (Line): A line from the manifest.

        Returns:
            Entry: An entry.

        Raises:
            ValueError: on parsing error.

        """
        part = [s.strip() for s in line.content.split(",")]
        if len(part) != 3:
            raise ValueError(
                _("Error: manifest line:{n}: " "must be: <relative_path>,<digest>,<size>").format(
                    n=line.number
                )
            )
        return Entry(relative_path=part[0], digest=part[1], size=int(part[2]))

    def __str__(self):
        """
        Returns a string representation of the Manifest Entry.

        Returns:
            str: format: "<relative_path>,<digest>,<size>"

        """
        fields = [self.relative_path, self.digest]
        if isinstance(self.size, int):
            fields.append(str(self.size))
        return ",".join(fields)


class Manifest:
    """
    A file manifest.

    Describes files contained within the directory.

    Attributes:
        relative_path (str): An relative path to the manifest.

    """

    def __init__(self, relative_path):
        """
        Create a new Manifest.

        Args:
            relative_path (str): An relative path to the manifest.

        """
        self.relative_path = relative_path

    @staticmethod
    def parse(manifest_str):
        """
        Parse a manifest string and yield entries.

        Yields:
            Entry: for each line.

        """
        for n, line in enumerate(manifest_str.splitlines(), 1):
            line = line.strip()
            if not line:
                continue
            if line.startswith("#"):
                continue
            yield Entry.parse(Line(number=n, content=line))

    def read(self):
        """
        Read the file at `relative_path` and yield entries.

        Yields:
            Entry: for each line.

        """
        with open(self.relative_path) as fp:
            yield from Manifest.parse(fp.read())

    def write(self, entries):
        """
        Write the manifest.

        Args:
            entries (iterable): The entries to be written.

        """
        with open(self.relative_path, "w+") as fp:
            for entry in entries:
                line = str(entry)
                fp.write(line)
                fp.write("\n")



from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase

from pulp_file.app.serializers import FileContentSerializer
from pulp_file.app.models import FileContent

from pulpcore.plugin.models import Artifact


class TestFileContentSerializer(TestCase):
    """Test FileContentSerializer."""

    def setUp(self):
        """Set up the FileContentSerializer tests."""
        self.artifact = Artifact.objects.create(
            sha1="cf6121b0425c2f2e3a2fcfe6f402d59730eb5661",
            sha224="9a6297eb28d91fad5277c0833856031d0e940432ad807658bd2b60f4",
            sha256="c8ddb3dcf8da48278d57b0b94486832c66a8835316ccf7ca39e143cbfeb9184f",
            sha384="53a8a0cebcb7780ed7624790c9d9a4d09ba74b47270d397f5ed7bc1c46777a0fbe362aaf2bbe7f0966a350a12d76e28d",  # noqa
            sha512="a94a65f19b864d184a2a5e07fa29766f08c6d49b6f624b3dd3a36a98267b9137d9c35040b3e105448a869c23c2aec04c9e064e3555295c1b8de6515eed4da27d",  # noqa
            size=1024,
            file=SimpleUploadedFile("test_filename", b"test content"),
        )

    def test_valid_data(self):
        """Test that the FileContentSerializer accepts valid data."""
        data = {
            "artifact": "/pulp/api/v3/artifacts/{}/".format(self.artifact.pk),
            "relative_path": "foo",
        }
        serializer = FileContentSerializer(data=data)
        self.assertTrue(serializer.is_valid())

    def test_absolute_path_data(self):
        """Test that the FileContentSerializer does not accept data."""
        data = {
            "artifact": "/pulp/api/v3/artifacts/{}/".format(self.artifact.pk),
            "relative_path": "/foo",
        }
        serializer = FileContentSerializer(data=data)
        self.assertFalse(serializer.is_valid())

    def test_duplicate_data(self):
        """Test that the FileContentSerializer does not accept data."""
        FileContent.objects.create(relative_path="foo", digest=self.artifact.sha256)
        data = {
            "artifact": "/pulp/api/v3/artifacts/{}/".format(self.artifact.pk),
            "relative_path": "foo",
        }
        serializer = FileContentSerializer(data=data)
        self.assertFalse(serializer.is_valid())



# coding=utf-8
"""Constants for Pulp File plugin tests."""
from urllib.parse import urljoin

from pulp_smash import config
from pulp_smash.pulp3.constants import (
    API_DOCS_PATH,
    BASE_DISTRIBUTION_PATH,
    BASE_PUBLICATION_PATH,
    BASE_REMOTE_PATH,
    BASE_CONTENT_PATH,
    BASE_REPO_PATH,
)


PULP_FIXTURES_BASE_URL = config.get_config().get_fixtures_url()

API_SCHEMA_PATH = urljoin(API_DOCS_PATH, ".json?format=openapi")

FILE_CONTENT_NAME = "file.file"

FILE_CONTENT_PATH = urljoin(BASE_CONTENT_PATH, "file/files/")

FILE_DISTRIBUTION_PATH = urljoin(BASE_DISTRIBUTION_PATH, "file/file/")

FILE_REMOTE_PATH = urljoin(BASE_REMOTE_PATH, "file/file/")

FILE_REPO_PATH = urljoin(BASE_REPO_PATH, "file/file/")

FILE_PUBLICATION_PATH = urljoin(BASE_PUBLICATION_PATH, "file/file/")

FILE_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file/")
"""The URL to a file repository."""

FILE_FIXTURE_MANIFEST_URL = urljoin(FILE_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest."""

FILE_FIXTURE_COUNT = 3
"""The number of packages available at :data:`FILE_FIXTURE_URL`."""

FILE_FIXTURE_SUMMARY = {FILE_CONTENT_NAME: FILE_FIXTURE_COUNT}
"""The desired content summary after syncing :data:`FILE_FIXTURE_URL`."""

FILE_URL = urljoin(FILE_FIXTURE_URL, "1.iso")
"""The URL to an ISO file at :data:`FILE_FIXTURE_URL`."""

FILE_URL2 = urljoin(FILE_FIXTURE_URL, "2.iso")
"""The URL to another ISO file at :data:`FILE_FIXTURE_URL`."""

FILE2_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file2/")
"""The URL to a file repository."""

FILE2_FIXTURE_MANIFEST_URL = urljoin(FILE2_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest"""

FILE2_URL = urljoin(FILE2_FIXTURE_URL, "1.iso")
"""The URL to an ISO file at :data:`FILE2_FIXTURE_URL`."""

FILE_INVALID_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-invalid/")
"""The URL to an invalid file repository."""

FILE_INVALID_MANIFEST_URL = urljoin(FILE_INVALID_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to an invalid file repository."""

FILE_LARGE_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-large/")
"""The URL to a file repository containing a large number of files."""

FILE_LARGE_FIXTURE_MANIFEST_URL = urljoin(FILE_LARGE_FIXTURE_URL, "PULP_MANIFEST")
"""The URL to a file repository manifest."""

FILE_PERFORMANCE_FIXTURE_URL = urljoin(PULP_FIXTURES_BASE_URL, "file-perf/")
"""The URL to a file repository used for performance tests."""



from pulpcore.plugin.importexport import BaseContentResource
from pulp_file.app.models import FileContent


class FileContentResource(BaseContentResource):
    """
    Resource for import/export of file_filecontent entities
    """

    def set_up_queryset(self):
        """
        :return: FileContents specific to a specified repo-version.
        """
        return FileContent.objects.filter(pk__in=self.repo_version.content)

    class Meta:
        model = FileContent
        import_id_fields = model.natural_key_fields()


IMPORT_ORDER = [FileContentResource]



# coding=utf-8
"""Tests that perform actions over content unit."""
import unittest

from pulp_smash import utils
from pulp_smash.pulp3.bindings import monitor_task, PulpTaskError
from pulp_smash.pulp3.utils import gen_repo, delete_orphans

from pulp_file.tests.functional.utils import (
    gen_artifact,
    gen_file_client,
    gen_file_content_attrs,
    gen_file_content_upload_attrs,
    tasks,
    skip_if,
)
from pulp_file.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_file import (
    ContentFilesApi,
    RepositoriesFileApi,
    RepositoriesFileVersionsApi,
)


class ContentUnitTestCase(unittest.TestCase):
    """CRUD content unit.

    This test targets the following issues:

    * `Pulp #2872 <https://pulp.plan.io/issues/2872>`_
    * `Pulp #3445 <https://pulp.plan.io/issues/3445>`_
    * `Pulp Smash #870 <https://github.com/pulp/pulp-smash/issues/870>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variable."""
        delete_orphans()
        cls.content_unit = {}
        cls.file_content_api = ContentFilesApi(gen_file_client())
        cls.artifact = gen_artifact()

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        delete_orphans()

    def test_01_create_content_unit(self):
        """Create content unit."""
        attrs = gen_file_content_attrs(self.artifact)
        response = self.file_content_api.create(**attrs)
        created_resources = monitor_task(response.task).created_resources
        content_unit = self.file_content_api.read(created_resources[0])
        self.content_unit.update(content_unit.to_dict())
        for key, val in attrs.items():
            with self.subTest(key=key):
                self.assertEqual(self.content_unit[key], val)

    @skip_if(bool, "content_unit", False)
    def test_02_read_content_unit(self):
        """Read a content unit by its href."""
        content_unit = self.file_content_api.read(self.content_unit["pulp_href"]).to_dict()
        for key, val in self.content_unit.items():
            with self.subTest(key=key):
                self.assertEqual(content_unit[key], val)

    @skip_if(bool, "content_unit", False)
    def test_02_read_content_units(self):
        """Read a content unit by its relative_path."""
        page = self.file_content_api.list(relative_path=self.content_unit["relative_path"])
        self.assertEqual(len(page.results), 1)
        for key, val in self.content_unit.items():
            with self.subTest(key=key):
                self.assertEqual(page.results[0].to_dict()[key], val)

    @skip_if(bool, "content_unit", False)
    def test_03_partially_update(self):
        """Attempt to update a content unit using HTTP PATCH.

        This HTTP method is not supported and a HTTP exception is expected.
        """
        attrs = gen_file_content_attrs(self.artifact)
        with self.assertRaises(AttributeError) as exc:
            self.file_content_api.partial_update(self.content_unit["pulp_href"], attrs)
        error_message = "'ContentFilesApi' object has no attribute 'partial_update'"
        self.assertEqual(exc.exception.args[0], error_message)

    @skip_if(bool, "content_unit", False)
    def test_03_fully_update(self):
        """Attempt to update a content unit using HTTP PUT.

        This HTTP method is not supported and a HTTP exception is expected.
        """
        attrs = gen_file_content_attrs(self.artifact)
        with self.assertRaises(AttributeError) as exc:
            self.file_content_api.update(self.content_unit["pulp_href"], attrs)
        error_message = "'ContentFilesApi' object has no attribute 'update'"
        self.assertEqual(exc.exception.args[0], error_message)

    @skip_if(bool, "content_unit", False)
    def test_04_delete(self):
        """Attempt to delete a content unit using HTTP DELETE.

        This HTTP method is not supported and a HTTP exception is expected.
        """
        with self.assertRaises(AttributeError) as exc:
            self.file_content_api.delete(self.content_unit["pulp_href"])
        error_message = "'ContentFilesApi' object has no attribute 'delete'"
        self.assertEqual(exc.exception.args[0], error_message)


class ContentUnitUploadTestCase(unittest.TestCase):
    """CRUD content unit with upload feature.

    This test targets the following issue:

    `Pulp #5403 <https://pulp.plan.io/issues/5403>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variable."""
        delete_orphans()
        cls.content_unit = {}
        cls.file_content_api = ContentFilesApi(gen_file_client())
        cls.attrs = gen_file_content_upload_attrs()

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        delete_orphans()

    def test_01_create_content_unit(self):
        """Create content unit."""
        response = self.file_content_api.create(**self.attrs, file=__file__)
        created_resources = monitor_task(response.task).created_resources
        content_unit = self.file_content_api.read(created_resources[0])
        self.content_unit.update(content_unit.to_dict())
        for key, val in self.attrs.items():
            with self.subTest(key=key):
                self.assertEqual(self.content_unit[key], val)

    @skip_if(bool, "content_unit", False)
    def test_02_read_content_unit(self):
        """Read a content unit by its href."""
        content_unit = self.file_content_api.read(self.content_unit["pulp_href"]).to_dict()
        for key, val in self.content_unit.items():
            with self.subTest(key=key):
                self.assertEqual(content_unit[key], val)

    @skip_if(bool, "content_unit", False)
    def test_02_read_content_units(self):
        """Read a content unit by its relative_path."""
        page = self.file_content_api.list(relative_path=self.content_unit["relative_path"])
        self.assertEqual(len(page.results), 1)
        for key, val in self.content_unit.items():
            with self.subTest(key=key):
                self.assertEqual(page.results[0].to_dict()[key], val)

    @skip_if(bool, "content_unit", False)
    def test_03_fail_duplicate_content_unit(self):
        """Create content unit."""
        response = self.file_content_api.create(**self.attrs, file=__file__)
        with self.assertRaises(PulpTaskError) as cm:
            monitor_task(response.task)
        task = cm.exception.task.to_dict()
        self.assertEqual(task["state"], "failed")
        error_description = task["error"]["description"]
        for key in ("already", "relative", "path", "digest"):
            self.assertIn(key, error_description.lower(), task["error"])

    @skip_if(bool, "content_unit", False)
    def test_03_duplicate_content_unit(self):
        """Create content unit."""
        attrs = self.attrs.copy()
        attrs["relative_path"] = utils.uuid4()
        response = self.file_content_api.create(**attrs, file=__file__)
        monitor_task(response.task)


class DuplicateContentUnit(unittest.TestCase):
    """Attempt to create a duplicate content unit.

    This test targets the following issues:

    *  `Pulp #4125 <https://pulp.plan.io/issue/4125>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.file_content_api = ContentFilesApi(gen_file_client())

    @classmethod
    def tearDownClass(cls):
        """Clean created resources."""
        delete_orphans()

    def test_raise_error(self):
        """Create a duplicate content unit using same relative_path.

        Artifacts are unique by ``relative_path`` and ``file``.

        In order to raise an HTTP error, the same ``artifact`` and the same
        ``relative_path`` should be used.
        """
        delete_orphans()
        artifact = gen_artifact()
        attrs = gen_file_content_attrs(artifact)

        # create first content unit.
        response = self.file_content_api.create(**attrs)
        monitor_task(response.task)

        # using the same attrs used to create the first content unit.
        response = self.file_content_api.create(**attrs)
        with self.assertRaises(PulpTaskError) as cm:
            monitor_task(response.task)
        error = cm.exception.task.to_dict()["error"]
        for key in ("already", "relative", "path", "digest"):
            self.assertIn(key, error["description"].lower(), error)

    def test_non_error(self):
        """Create a duplicate content unit with different relative_path.

        Artifacts are unique by ``relative_path`` and ``file``.

        In order to avoid an HTTP error, use the same ``artifact`` and
        different ``relative_path``.
        """
        delete_orphans()
        artifact = gen_artifact()

        # create first content unit.
        response = self.file_content_api.create(**gen_file_content_attrs(artifact))
        monitor_task(response.task)

        # create second content unit.
        response = self.file_content_api.create(**gen_file_content_attrs(artifact))
        monitor_task(response.task)
        task = tasks.read(response.task)
        self.assertEqual(task.state, "completed")


class DuplicateRelativePathsInRepo(unittest.TestCase):
    """Associate different Content units with the same ``relative_path`` in one RepositoryVersion.

    This test targets the following issues:

    *  `Pulp #4028 <https://pulp.plan.io/issue/4028>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_file_client()

    @classmethod
    def tearDownClass(cls):
        """Clean created resources."""
        delete_orphans()

    def test_second_unit_replaces_the_first(self):
        """Create a duplicate content unit with different ``artifacts`` and same ``relative_path``.

        Artifacts are unique by ``relative_path`` and ``file``.
        """
        delete_orphans()
        content_api = ContentFilesApi(self.client)
        repo_api = RepositoriesFileApi(self.client)
        versions_api = RepositoriesFileVersionsApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        artifact = gen_artifact()

        # create first content unit.
        content_attrs = gen_file_content_attrs(artifact)
        content_attrs["repository"] = repo.pulp_href
        response = content_api.create(**content_attrs)
        monitor_task(response.task)

        artifact = gen_artifact(file=__file__)

        # create second content unit.
        second_content_attrs = gen_file_content_attrs(artifact)
        second_content_attrs["repository"] = repo.pulp_href
        second_content_attrs["relative_path"] = content_attrs["relative_path"]

        response = content_api.create(**second_content_attrs)
        monitor_task(response.task)

        repo_latest_version = versions_api.read(repo_api.read(repo.pulp_href).latest_version_href)

        self.assertEqual(repo_latest_version.content_summary.present["file.file"]["count"], 1)

    def test_second_unit_raises_error(self):
        """Create a duplicate content unit with different ``artifacts`` and same ``relative_path``.

        Artifacts are unique by ``relative_path`` and ``file``.
        """
        delete_orphans()
        content_api = ContentFilesApi(self.client)
        repo_api = RepositoriesFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        artifact = gen_artifact()

        # create first content unit.
        content_attrs = gen_file_content_attrs(artifact)
        response = content_api.create(**content_attrs)
        monitor_task(response.task)

        artifact = gen_artifact(file=__file__)

        # create second content unit.
        second_content_attrs = gen_file_content_attrs(artifact)
        second_content_attrs["relative_path"] = content_attrs["relative_path"]
        response = content_api.create(**second_content_attrs)
        monitor_task(response.task)

        data = {"add_content_units": [c.pulp_href for c in content_api.list().results]}
        response = repo_api.modify(repo.pulp_href, data)
        with self.assertRaises(PulpTaskError) as cm:
            monitor_task(response.task)
        task = cm.exception.task.to_dict()

        error_message = (
            "Cannot create repository version. "
            "More than one file.file content with "
            "the duplicate values for relative_path."
        )
        self.assertEqual(task["error"]["description"], error_message)



from django.http import Http404
from django_filters import CharFilter
from drf_spectacular.utils import extend_schema
from rest_framework.decorators import action

from pulpcore.plugin.actions import ModifyRepositoryActionMixin
from pulpcore.plugin.serializers import (
    AsyncOperationResponseSerializer,
    PublicationExportSerializer,
    RepositorySyncURLSerializer,
)
from pulpcore.plugin.tasking import enqueue_with_reservation, fs_publication_export
from pulpcore.plugin.viewsets import (
    BaseDistributionViewSet,
    ContentFilter,
    ExporterViewSet,
    ExportViewSet,
    OperationPostponedResponse,
    PublicationViewSet,
    RemoteViewSet,
    RepositoryViewSet,
    RepositoryVersionViewSet,
    SingleArtifactContentUploadViewSet,
)

from . import tasks
from .models import (
    FileContent,
    FileDistribution,
    FileFilesystemExporter,
    FileRemote,
    FileRepository,
    FilePublication,
)
from .serializers import (
    FileContentSerializer,
    FileDistributionSerializer,
    FileFilesystemExporterSerializer,
    FileRemoteSerializer,
    FileRepositorySerializer,
    FilePublicationSerializer,
)


class FileContentFilter(ContentFilter):
    """
    FilterSet for FileContent.
    """

    sha256 = CharFilter(field_name="_artifacts__sha256")

    class Meta:
        model = FileContent
        fields = ["relative_path", "sha256"]


class FileContentViewSet(SingleArtifactContentUploadViewSet):
    """
    <!-- User-facing documentation, rendered as html-->
    FileContent represents a single file and its metadata, which can be added and removed from
    repositories.
    """

    endpoint_name = "files"
    queryset = FileContent.objects.prefetch_related("_artifacts")
    serializer_class = FileContentSerializer
    filterset_class = FileContentFilter


class FileRepositoryViewSet(RepositoryViewSet, ModifyRepositoryActionMixin):
    """
    <!-- User-facing documentation, rendered as html-->
    FileRepository represents a single file repository, to which content can be synced, added,
    or removed.
    """

    endpoint_name = "file"
    queryset = FileRepository.objects.all()
    serializer_class = FileRepositorySerializer

    @extend_schema(
        description="Trigger an asynchronous task to sync file content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=RepositorySyncURLSerializer)
    def sync(self, request, pk):
        """
        Synchronizes a repository.

        The ``repository`` field has to be provided.
        """
        serializer = RepositorySyncURLSerializer(
            data=request.data, context={"request": request, "repository_pk": pk}
        )
        serializer.is_valid(raise_exception=True)

        repository = self.get_object()
        remote = serializer.validated_data.get("remote", repository.remote)

        mirror = serializer.validated_data.get("mirror", False)
        result = enqueue_with_reservation(
            tasks.synchronize,
            [repository, remote],
            kwargs={"remote_pk": remote.pk, "repository_pk": repository.pk, "mirror": mirror},
        )
        return OperationPostponedResponse(result, request)


class FileRepositoryVersionViewSet(RepositoryVersionViewSet):
    """
    <!-- User-facing documentation, rendered as html-->
    FileRepositoryVersion represents a single file repository version.
    """

    parent_viewset = FileRepositoryViewSet


class FileRemoteViewSet(RemoteViewSet):
    """
    <!-- User-facing documentation, rendered as html-->
    FileRemote represents an external source of <a href="#operation/content_file_files_list">File
    Content</a>.  The target url of a FileRemote must contain a file manifest, which contains the
    metadata for all files at the source.
    """

    endpoint_name = "file"
    queryset = FileRemote.objects.all()
    serializer_class = FileRemoteSerializer


class FilePublicationViewSet(PublicationViewSet):
    """
    <!-- User-facing documentation, rendered as html-->
    A FilePublication contains metadata about all the <a
    href="#operation/content_file_files_list">File Content</a> in a particular <a
    href="href="#tag/repositories:-file-versions">File Repository Version.</a>
    Once a FilePublication has been created, it can be hosted using the
    <a href="#operation/distributions_file_file_list">File Distribution API.</a>
    """

    endpoint_name = "file"
    queryset = FilePublication.objects.exclude(complete=False)
    serializer_class = FilePublicationSerializer

    @extend_schema(
        description="Trigger an asynchronous task to publish file content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """
        Publishes a repository.

        Either the ``repository`` or the ``repository_version`` fields can
        be provided but not both at the same time.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        repository_version = serializer.validated_data.get("repository_version")
        manifest = serializer.validated_data.get("manifest")

        result = enqueue_with_reservation(
            tasks.publish,
            [repository_version.repository],
            kwargs={"repository_version_pk": str(repository_version.pk), "manifest": str(manifest)},
        )
        return OperationPostponedResponse(result, request)


class FileDistributionViewSet(BaseDistributionViewSet):
    """
    <!-- User-facing documentation, rendered as html-->
    FileDistributions host <a href="#operation/publications_file_file_list">File
    Publications</a> which makes the metadata and the referenced <a
    href="#operation/content_file_files_list">File Content</a> available to HTTP
    clients. Additionally, a FileDistribution with an associated FilePublication can be the target
    url of a <a href="#operation/remotes_file_file_list">File Remote</a> , allowing
    another instance of Pulp to sync the content.
    """

    endpoint_name = "file"
    queryset = FileDistribution.objects.all()
    serializer_class = FileDistributionSerializer


class FileFilesystemExporterViewSet(ExporterViewSet):
    """
    FilesystemExporters export content from a publication to a path on the file system.

    WARNING: This feature is provided as a tech preview and may change in the future. Backwards
    compatibility is not guaranteed.
    """

    endpoint_name = "filesystem"
    queryset = FileFilesystemExporter.objects.all()
    serializer_class = FileFilesystemExporterSerializer


class FileFilesystemExportViewSet(ExportViewSet):
    """
    FilesystemExports provide a history of previous exports.
    """

    parent_viewset = FileFilesystemExporterViewSet

    @extend_schema(
        request=PublicationExportSerializer,
        description="Trigger an asynchronous task to export a file publication.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request, exporter_pk):
        """
        Export a publication to the file system.

        The ``repository`` field has to be provided.
        """
        try:
            exporter = FileFilesystemExporter.objects.get(pk=exporter_pk)
        except FileFilesystemExporter.DoesNotExist:
            raise Http404

        serializer = PublicationExportSerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)
        publication = serializer.validated_data.get("publication")
        result = enqueue_with_reservation(
            fs_publication_export,
            [publication, exporter],
            kwargs={"exporter_pk": exporter.pk, "publication_pk": publication.pk},
        )
        return OperationPostponedResponse(result, request)



# Generated by Django 2.2.4 on 2019-08-09 17:03

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('file', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='filecontent',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='filedistribution',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='filepublication',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterModelOptions(
            name='fileremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s'},
        ),
        migrations.AlterField(
            model_name='filecontent',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='file_filecontent', serialize=False, to='core.Content'),
        ),
        migrations.AlterField(
            model_name='filedistribution',
            name='basedistribution_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='file_filedistribution', serialize=False, to='core.BaseDistribution'),
        ),
        migrations.AlterField(
            model_name='filedistribution',
            name='publication',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='file_filedistribution', to='core.Publication'),
        ),
        migrations.AlterField(
            model_name='filepublication',
            name='publication_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='file_filepublication', serialize=False, to='core.Publication'),
        ),
        migrations.AlterField(
            model_name='fileremote',
            name='remote_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='file_fileremote', serialize=False, to='core.Remote'),
        ),
    ]



# coding=utf-8
"""Tests for Pulp`s download policies."""
from random import choice
import unittest

from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.constants import ON_DEMAND_DOWNLOAD_POLICIES
from pulp_smash.pulp3.utils import (
    delete_orphans,
    gen_repo,
    get_added_content_summary,
    get_content_summary,
)

from pulp_file.tests.functional.constants import (
    FILE_FIXTURE_COUNT,
    FILE_FIXTURE_SUMMARY,
)
from pulp_file.tests.functional.utils import (
    core_client,
    gen_file_client,
    gen_file_remote,
    skip_if,
)
from pulp_file.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulpcore import ArtifactsApi
from pulpcore.client.pulp_file import (
    ContentFilesApi,
    FileFilePublication,
    PublicationsFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
    RemotesFileApi,
)


class SyncPublishDownloadPolicyTestCase(unittest.TestCase):
    """Sync/Publish a repository with different download policies.

    This test targets the following issues:

    `Pulp #4126 <https://pulp.plan.io/issues/4126>`_
    `Pulp #4418 <https://pulp.plan.io/issues/4418>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_file_client()
        cls.DP_ON_DEMAND = "on_demand" in ON_DEMAND_DOWNLOAD_POLICIES
        cls.DP_STREAMED = "streamed" in ON_DEMAND_DOWNLOAD_POLICIES

    @skip_if(bool, "DP_ON_DEMAND", False)
    def test_on_demand(self):
        """Sync with ``on_demand`` download policy. See :meth:`do_sync`."""
        self.do_sync("on_demand")
        self.do_publish("on_demand")

    @skip_if(bool, "DP_STREAMED", False)
    def test_streamed(self):
        """Sync with ``streamend`` download policy.  See :meth:`do_sync`."""
        self.do_sync("streamed")
        self.do_publish("streamed")

    def do_sync(self, download_policy):
        """Sync repositories with the different ``download_policy``.

        Do the following:

        1. Create a repository, and a remote.
        2. Assert that repository version is None.
        3. Sync the remote.
        4. Assert that repository version is not None.
        5. Assert that the correct number of possible units to be downloaded
           were shown.
        6. Sync the remote one more time in order to create another repository
           version.
        7. Assert that repository version is the same as the previous one.
        8. Assert that the same number of units are shown, and after the
           second sync no extra units should be shown, since the same remote
           was synced again.
        """
        # delete orphans to assure that no content units are present on the
        # file system
        delete_orphans()
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote(**{"policy": download_policy})
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        self.assertIsNotNone(repo.latest_version_href)
        self.assertDictEqual(get_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)
        self.assertDictEqual(get_added_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)

        # Sync the repository again.
        latest_version_href = repo.latest_version_href
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        self.assertEqual(latest_version_href, repo.latest_version_href)
        self.assertDictEqual(get_content_summary(repo.to_dict()), FILE_FIXTURE_SUMMARY)

    def do_publish(self, download_policy):
        """Publish repository synced with lazy download policy."""
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)
        publications = PublicationsFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote(policy=download_policy)
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        publish_data = FileFilePublication(repository=repo.pulp_href)
        publish_response = publications.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        self.addCleanup(publications.delete, publication_href)
        publication = publications.read(publication_href)
        self.assertIsNotNone(publication.repository_version, publication)


class LazySyncedContentAccessTestCase(unittest.TestCase):
    """Verify that lazy synced content can be acessed using content endpoint.

    Assert that one acessing lazy synced content using the content endpoint,
    e.g. ``http://localhost/pulp/api/v3/content/file/files`` will not raise an
    HTTP exception.

    This test targets the following issue:

    `Pulp #4463 <https://pulp.plan.io/issues/4463>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_file_client()

    def test_on_demand(self):
        """Test ``on_demand``. See :meth:`do_test`."""
        self.do_test("on_demand")

    def test_streamed(self):
        """Test ``streamed``. See :meth:`do_test`."""
        self.do_test("streamed")

    def do_test(self, policy):
        """Access lazy synced content on using content endpoint."""
        # delete orphans to assure that no content units are present on the
        # file system
        delete_orphans()
        content_api = ContentFilesApi(self.client)
        repo_api = RepositoriesFileApi(self.client)
        remote_api = RemotesFileApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote(**{"policy": policy})
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        # Assert that no HTTP error was raised.
        # Assert that the number of units present is according to the synced
        # feed.
        content = content_api.list().to_dict()["results"]
        self.assertEqual(len(content), FILE_FIXTURE_COUNT, content)


class SwitchDownloadPolicyTestCase(unittest.TestCase):
    """Perform a lazy sync, and change to immediate to force download.

    Perform an immediate sync to download artifacts for content units that
    are already created.

    This test case targets the following issue:

    * `Pulp #4467 <https://pulp.plan.io/issues/4467>`_
    """

    def test_all(self):
        """Perform a lazy sync and change to immeditae to force download."""
        # delete orphans to assure that no content units are present on the
        # file system
        delete_orphans()
        client = gen_file_client()
        artifacts_api = ArtifactsApi(core_client)
        repo_api = RepositoriesFileApi(client)
        remote_api = RemotesFileApi(client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote(policy=choice(ON_DEMAND_DOWNLOAD_POLICIES))
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository using a lazy download policy
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        artifacts = artifacts_api.list().to_dict()["results"]
        self.assertEqual(len(artifacts), 0, artifacts)

        # Update the policy to immediate
        response = remote_api.partial_update(remote.pulp_href, {"policy": "immediate"})
        monitor_task(response.task)
        remote = remote_api.read(remote.pulp_href)
        self.assertEqual(remote.policy, "immediate")

        # Sync using immediate download policy
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        # Assert that missing artifacts are downloaded
        artifacts = artifacts_api.list().to_dict()["results"]
        self.assertEqual(len(artifacts), FILE_FIXTURE_COUNT, artifacts)



# Generated by Django 2.2.1 on 2019-05-30 19:40

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilePublication',
            fields=[
                ('publication_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Publication')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.publication',),
        ),
        migrations.CreateModel(
            name='FileRemote',
            fields=[
                ('remote_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Remote')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.remote',),
        ),
        migrations.CreateModel(
            name='FileDistribution',
            fields=[
                ('basedistribution_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.BaseDistribution')),
                ('publication', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Publication')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.basedistribution',),
        ),
        migrations.CreateModel(
            name='FileContent',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Content')),
                ('relative_path', models.CharField(max_length=255)),
                ('digest', models.CharField(max_length=64)),
            ],
            options={
                'unique_together': {('relative_path', 'digest')},
            },
            bases=('core.content',),
        ),
    ]



import logging
import os

from gettext import gettext as _
from urllib.parse import urlparse, urlunparse

from pulpcore.plugin.models import Artifact, ProgressReport, Remote
from pulpcore.plugin.stages import (
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    Stage,
)

from pulp_file.app.models import FileContent, FileRemote, FileRepository
from pulp_file.manifest import Manifest


log = logging.getLogger(__name__)


def synchronize(remote_pk, repository_pk, mirror):
    """
    Sync content from the remote repository.

    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.

    Raises:
        ValueError: If the remote does not specify a URL to sync.

    """
    remote = FileRemote.objects.get(pk=remote_pk)
    repository = FileRepository.objects.get(pk=repository_pk)

    if not remote.url:
        raise ValueError(_("A remote must have a url specified to synchronize."))

    first_stage = FileFirstStage(remote)
    dv = DeclarativeVersion(first_stage, repository, mirror=mirror)
    dv.create()


class FileFirstStage(Stage):
    """
    The first stage of a pulp_file sync pipeline.
    """

    def __init__(self, remote):
        """
        The first stage of a pulp_file sync pipeline.

        Args:
            remote (FileRemote): The remote data to be used when syncing

        """
        super().__init__()
        self.remote = remote

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the Manifest data.
        """
        deferred_download = self.remote.policy != Remote.IMMEDIATE  # Interpret download policy
        with ProgressReport(message="Downloading Metadata", code="downloading.metadata") as pb:
            parsed_url = urlparse(self.remote.url)
            root_dir = os.path.dirname(parsed_url.path)
            downloader = self.remote.get_downloader(url=self.remote.url)
            result = await downloader.run()
            pb.increment()

        with ProgressReport(message="Parsing Metadata Lines", code="parsing.metadata") as pb:
            manifest = Manifest(result.path)
            entries = list(manifest.read())

            pb.total = len(entries)
            pb.save()

            for entry in entries:
                path = os.path.join(root_dir, entry.relative_path)
                url = urlunparse(parsed_url._replace(path=path))
                file = FileContent(relative_path=entry.relative_path, digest=entry.digest)
                artifact = Artifact(size=entry.size, sha256=entry.digest)
                da = DeclarativeArtifact(
                    artifact=artifact,
                    url=url,
                    relative_path=entry.relative_path,
                    remote=self.remote,
                    deferred_download=deferred_download,
                )
                dc = DeclarativeContent(content=file, d_artifacts=[da])
                pb.increment()
                await self.put(dc)
